@@ -113,24 +113,24 @@ def assemble_pushdata(pushdata: bytes) -> bytes:
     return get_pushdata_code(len(pushdata)) + pushdata
 
 
-def deserialize_ecdsa_der(der: bytes) -> Tuple[int, int]:
+def deserialize_ecdsa_der(signature: bytes) -> Tuple[int, int]:
     """
-    deserialize ECDSA signature from bitcoin DER to (r, s)
+    deserialize ECDSA signature from bitcoin strict DER to (r, s)
     """
     try:
-        assert der[0] == 0x30
-        assert int(der[1]) == len(der) - 2
+        assert signature[0] == 0x30
+        assert int(signature[1]) == len(signature) - 2
         # r
-        assert der[2] == 0x02
-        r_len = int(der[3])
-        r = int.from_bytes(der[4: 4 + r_len], 'big')
+        assert signature[2] == 0x02
+        r_len = int(signature[3])
+        r = int.from_bytes(signature[4: 4 + r_len], 'big')
         # s
-        assert der[4 + r_len] == 0x02
-        s_len = int(der[5 + r_len])
-        s = int.from_bytes(der[-s_len:], 'big')
+        assert signature[4 + r_len] == 0x02
+        s_len = int(signature[5 + r_len])
+        s = int.from_bytes(signature[-s_len:], 'big')
         return r, s
     except Exception:
-        raise ValueError(f'invalid DER encoded {der.hex()}')
+        raise ValueError(f'invalid DER encoded {signature.hex()}')
 
 
 def serialize_ecdsa_der(signature: Tuple[int, int]) -> bytes:
@@ -154,6 +154,25 @@ def serialize_ecdsa_der(signature: Tuple[int, int]) -> bytes:
     return bytes([0x30, len(serialized)]) + serialized
 
 
+def deserialize_ecdsa_recoverable(signature: bytes) -> Tuple[int, int, int]:
+    """
+    deserialize recoverable ECDSA signature from bytes to (r, s, recovery_id)
+    """
+    assert len(signature) == 65, 'invalid length of recoverable ECDSA signature'
+    recovery_id = signature[-1]
+    assert 0 <= recovery_id <= 3, f'invalid recovery id {recovery_id}'
+    return int.from_bytes(signature[:32], 'big'), int.from_bytes(signature[32:-1], 'big'), recovery_id
+
+
+def serialize_ecdsa_recoverable(signature: Tuple[int, int, int]) -> bytes:
+    """
+    serialize recoverable ECDSA signature from (r, s, recovery_id) to bytes
+    """
+    r, s, recovery_id = signature
+    assert 0 <= recovery_id < 4, f'invalid recovery id {recovery_id}'
+    return r.to_bytes(NUMBER_BYTE_LENGTH, 'big') + s.to_bytes(NUMBER_BYTE_LENGTH, 'big') + recovery_id.to_bytes(1, 'big')
+
+
 def serialize_text(text: str) -> bytes:
     """
     serialize plain text to bytes in format: varint_length + text.utf-8
@@ -169,30 +188,32 @@ def text_digest(text: str) -> bytes:
     return serialize_text('Bitcoin Signed Message:\n') + serialize_text(text)
 
 
-def serialize_ecdsa_recoverable(signature: Tuple[int, int, int], compressed: bool = True) -> str:
-    """serialize recoverable ECDSA signature (recovery_id, r, s), compressed is True if used compressed public key
-    :returns: serialized recoverable signature formatted in base64
+def stringify_ecdsa_recoverable(signature: bytes, compressed: bool = True) -> str:
+    """stringify serialize recoverable ECDSA signature
+    :param signature: serialized recoverable ECDSA signature in format "r (32 bytes) + s (32 bytes) + recovery_id (1 byte)"
+    :param compressed: True if used compressed public key
+    :returns: stringified recoverable signature formatted in base64
     """
-    recovery_id, r, s = signature
+    r, s, recovery_id = deserialize_ecdsa_recoverable(signature)
     prefix: int = 27 + recovery_id + (4 if compressed else 0)
-    signature: bytes = prefix.to_bytes(1, 'big') + r.to_bytes(NUMBER_BYTE_LENGTH, 'big') + s.to_bytes(NUMBER_BYTE_LENGTH, 'big')
+    signature: bytes = prefix.to_bytes(1, 'big') + signature[:-1]
     return b64encode(signature).decode('ascii')
 
 
-def deserialize_ecdsa_recoverable(signature: str) -> Tuple[Tuple[int, int, int], bool]:
+def unstringify_ecdsa_recoverable(signature: str) -> Tuple[bytes, bool]:
     """
-    :returns: ((recovery_id, r, s), used_compressed_public_key)
+    :returns: (serialized_recoverable_signature, used_compressed_public_key)
     """
-    signature_bytes = b64decode(signature)
-    assert len(signature_bytes) == 65, f'invalid recoverable ECDSA signature {signature}'
-    prefix, r, s = signature_bytes[0], int.from_bytes(signature_bytes[1:33], 'big'), int.from_bytes(signature_bytes[33:], 'big')
+    serialized = b64decode(signature)
+    assert len(serialized) == 65, 'invalid length of recoverable ECDSA signature'
+    prefix = serialized[0]
     assert 27 <= prefix < 35, f'invalid recoverable ECDSA signature prefix {prefix}'
     compressed = False
     if prefix >= 31:
         compressed = True
         prefix -= 4
     recovery_id = prefix - 27
-    return (recovery_id, r, s), compressed
+    return serialized[1:] + recovery_id.to_bytes(1, 'big'), compressed
 
 
 def bytes_to_bits(value: Union[str, bytes]) -> str:
